// $Id: example.cpp,v 1.12 2005/03/14 18:05:32 ejankows Exp $


//  +---------------------------------------------------------+
//  |                                                         |
//  |  OPTIMAL JET FINDER v1.0                                |
//  |                                                         |
//  |  (C) 2005 S. Chumakov, E. Jankowski and F. V. Tkachov   |
//  |                                                         |
//  |  Permission is granted for anyone to use this software  |       
//  |  for any purpose on any computer. The authors are not   |
//  |  responsible for the consequences of such use.          |        
//  |  This software is meant to be freely distributed.       |
//  |                                                         |
//  +---------------------------------------------------------+


// A very simple example of using Optimal Jet Finder
// INPUT: a list of (px, py, pz) from WW->qqqq event; file: inputWW.dat

// References:
// [1] D. Yu. Grigoriev, E. Jankowski, F. V. Tkachov,
//     Comput. Phys. Commun. 155, 42 (2003) [hep-ph/0301226].
// [2] S. Chumakov, E. Jankowski, F. V. Tkachov, 
//     the accompanying Comput. Phys. Commun. paper [paper.pdf].

#include "OJFKinematics.h"
#include "OJFJets.h"
#include "OJFSearch.h"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
using namespace std;
using namespace OptimalJetFinder;


int main() {

  //input data

  ifstream in( "inputWW.dat" );
    
  //create a new event
  Event *P = new Event( sphere );
  //use "cylinder" instead of "sphere" for cylindrical kinematics

  double px, py, pz;
  while( in>>px>>py>>pz ) {
    P->AddParticleRaw( px, py, pz ); //input a particle
  }

  in.close();

  //input data ends


  //normalize input momenta so that the sum of input energies = 1 
  //(or the sum transverse energies for cylindrical kinematics = 1)
  P->Normalize();

  cout << P->GetNumber() << " particles in the event." << endl;

  //set the seed for the random number generator
  OJFRandom::SetSeed( 13 );

  double radius = 1.0; // R parameter of eq. (20) in reference [1]
  unsigned ntries = 3; // number of tries

  //new jet search created
  JetSearch* js = new JetSearch( P, radius, ntries );

  //find jets for a given value of Omega_cut
  unsigned njets = js->FindJetsForOmegaCut(0.05);
  if( njets == 0 ) { cout << "Jets lost." << endl; exit(1); }

  //alternatively, find jet configuration for a fixed number of jets
  //bool success = js->FindJetsForFixedNJets(4);
  //if (  ! success  ) { cout << "Jets lost." << endl; exit(1); } 

  //get the jet configuration
  Jets* Q = js->GetJets();

  //display the number of jets
  // and parameters /Omega, Y, Esoft/ of the jet configuration
  cout << Q->GetNumber() << " jets found." << endl;
  cout << "Omega: " << Q->GetOmega() << ",   "
       << "Y: " << Q->GetY() << ",   "
       << "Esoft (normalized): " << Q->GetESoft() << "." << endl;

  //display the details of the jets
  cout << "The details of the jets (E px py pz):" << endl;
  
  Jet* jet = Q->GetFirst(); 
  while( jet ) {
    cout << setw( 10 ) << jet->GetE() << "  "
	 << setw( 10 ) << jet->GetPx() << "  "
	 << setw( 10 ) << jet->GetPy() << "  "
	 << setw( 10 ) << jet->GetPz() << endl;
    jet = jet->GetNext();
  }

  //the user is responsible for deleting what they created themselves with new
  delete P;
  delete js;

}
